# -*- coding: utf-8 -*-
# Python library for Remember The Milk API

__author__ = 'Sridhar Ratnakumar <http://nearfar.org/>'
__all__ = (
    'API',
    'API_RESPONSE',
    'createRTM',
    'set_log_level',
)


import warnings
import urllib
import logging
from hashlib import md5

# rtm constant
from consts import *

try:
    import json
except ImportError:
    import simplejson as json

warnings.simplefilter('default', ImportWarning)

logging.basicConfig()
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


class RTMError(Exception): pass

class RTMAPIError(RTMError): pass

class AuthStateMachine(object):

    class NoData(RTMError): pass

    def __init__(self, states):
        self.states = states
        self.data = {}

    def dataReceived(self, state, datum):
        if state not in self.states:
            raise RTMError('Invalid state <%s>' % state)
        self.data[state] = datum

    def get(self, state):
        if state in self.data:
            return self.data[state]
        else:
            raise AuthStateMachine.NoData('No data for <%s>' % state)


class RTM(object):

    def __init__(self, apiKey, secret, token=None):
        self.apiKey = apiKey
        self.secret = secret
        self.authInfo = AuthStateMachine(['frob', 'token'])

        # this enables one to do 'rtm.tasks.getList()', for example
        for prefix, methods in API.items():
            setattr(self, prefix,
                    RTMAPICategory(self, prefix, methods))

        if token:
            self.authInfo.dataReceived('token', token)

    def _sign(self, params):
        "Sign the parameters with MD5 hash"
        pairs = ''.join(['%s%s' % (k, v) for k, v in sortedItems(params)])
        return md5(self.secret + pairs).hexdigest()

    def get(self, **params):
        "Get the XML response for the passed `params`."
        params['api_key'] = self.apiKey
        params['format'] = 'json'
        params['api_sig'] = self._sign(params)

        data = openURL(SERVICE_URL, params).read()

        LOG.debug("JSON response: \n%s" % data)

        d = dottedDict('ROOT', json.loads(data))
        rsp = d.rsp

        if rsp.stat == 'fail':
            raise RTMAPIError('API call failed - %s (%s)' % (
                rsp.err.msg, rsp.err.code))
        else:
            return rsp

    def getNewFrob(self):
        rsp = self.get(method='rtm.auth.getFrob')
        self.authInfo.dataReceived('frob', rsp.frob)
        return rsp.frob

    def getAuthURL(self):
        try:
            frob = self.authInfo.get('frob')
        except AuthStateMachine.NoData:
            frob = self.getNewFrob()

        params = {
            'api_key': self.apiKey,
            'perms': 'delete',
            'frob': frob
            }
        params['api_sig'] = self._sign(params)
        return AUTH_SERVICE_URL + '?' + urllib.urlencode(params)

    def getToken(self):
        frob = self.authInfo.get('frob')
        rsp = self.get(method='rtm.auth.getToken', frob=frob)
        self.authInfo.dataReceived('token', rsp.auth.token)
        return rsp.auth.token

class RTMAPICategory:
    "See the `API` structure and `RTM.__init__`"

    def __init__(self, rtm, prefix, methods):
        self.rtm = rtm
        self.prefix = prefix
        self.methods = methods

    def __getattr__(self, attr):
        if attr in self.methods:
            rargs, oargs = self.methods[attr]
            if self.prefix == 'tasksNotes':
                aname = 'rtm.tasks.notes.%s' % attr
            else:
                aname = 'rtm.%s.%s' % (self.prefix, attr)
            return lambda **params: self.callMethod(
                aname, rargs, oargs, **params)
        else:
            raise AttributeError('No such attribute: %s' % attr)

    def callMethod(self, aname, rargs, oargs, **params):
        # Sanity checks
        for requiredArg in rargs:
            if requiredArg not in params:
                raise TypeError('Required parameter (%s) missing' % (
                                requiredArg))

        for param in params:
            if param not in rargs + oargs:
                warnings.warn('Invalid parameter (%s)' % param)

        token = self.rtm.authInfo.get('token')
        if params.get('auth_token'):
            # auth.checkToken api needs auth_token parameter
            token = params.pop('auth_token')

        return self.rtm.get(method=aname, auth_token=token, **params)


# Utility functions

def sortedItems(dictionary):
    """ Return a list of (key, value) sorted based on keys
    >>> list(sortedItems({'a': 1, 1: 55, 'key': 7}))
    [(1, 55), ('a', 1), ('key', 7)]
    """
    keys = dictionary.keys()
    keys.sort()
    for key in keys:
        yield key, dictionary[key]

def openURL(url, queryArgs=None):
    """
    >>> q = {'query': 'test'}
    >>> r = openURL('http://www.rememberthemilk.com/', q)
    >>> r.geturl()
    'http://www.rememberthemilk.com/?query=test'
    """
    if queryArgs:
        url = url + '?' + urllib.urlencode(queryArgs)
    LOG.debug("URL> %s", url)
    return urllib.urlopen(url)

class dottedDict(object):
    """Make dictionary items accessible via the object-dot notation."""

    def __init__(self, name, dictionary):
        self._name = name

        if type(dictionary) is dict:
            for key, value in dictionary.items():
                if type(value) is dict:
                    value = dottedDict(key, value)
                elif type(value) in (list, tuple) and key != 'tag':
                    value = [dottedDict('%s_%d' % (key, i), item)
                             for i, item in indexed(value)]
                setattr(self, key, value)
        else:
            raise ValueError('not a dict: %s' % dictionary)

    def __repr__(self):
        children = [c for c in dir(self) if not c.startswith('_')]
        return 'dotted <%s> : %s' % (
            self._name,
            ', '.join(children))

def indexed(seq):
    """
    >>> list(indexed(['a', 'b', 'c']))
    [(0, 'a'), (1, 'b'), (2, 'c')]
    """
    index = 0
    for item in seq:
        yield index, item
        index += 1

def createRTM(apiKey, secret, token=None):
    rtm = RTM(apiKey, secret, token)

    if token is None:
        print 'No token found'
        print 'Give me access here:', rtm.getAuthURL()
        raw_input('Press enter once you gave access')
        print 'Note down this token for future use:', rtm.getToken()

    return rtm

def set_log_level(level):
    '''Sets the log level of the logger used by the module.

    >>> import rtm
    >>> import logging
    >>> rtm.set_log_level(logging.INFO)
    '''

    LOG.setLevel(level)



# -*- coding: utf-8 -*-

SERVICE_URL = 'http://api.rememberthemilk.com/services/rest/'
AUTH_SERVICE_URL = 'http://www.rememberthemilk.com/services/auth/'

# API spec
API = {
    'auth': {
        'checkToken':
            [('auth_token',), ()],
        'getFrob':
            [(), ()],
        'getToken':
            [('frob',), ()]
    },
    'contacts': {
        'add':
            [('timeline', 'contact'), ()],
        'delete':
            [('timeline', 'contact_id'), ()],
        'getList':
            [(), ()]
    },
    'groups': {
        'add':
            [('timeline', 'group'), ()],
        'addContact':
            [('timeline', 'group_id', 'contact_id'), ()],
        'delete':
            [('timeline', 'group_id'), ()],
        'getList':
            [(), ()],
        'removeContact':
            [('timeline', 'group_id', 'contact_id'), ()],
    },
    'lists': {
        'add':
            [('timeline', 'name',), ('filter',)],
        'archive':
            [('timeline', 'list_id'), ()],
        'delete':
            [('timeline', 'list_id'), ()],
        'getList':
            [(), ()],
        'setDefaultList':
            [('timeline'), ('list_id')],
        'setName':
            [('timeline', 'list_id', 'name'), ()],
        'unarchive':
            [('timeline',), ('list_id',)]
    },
    'locations': {
        'getList':
            [(), ()]
    },
    'reflection': {
        'getMethodInfo':
            [('method_name',), ()],
        'getMethods':
            [(), ()]
    },
    'settings': {
        'getList':
            [(), ()]
    },
    'tasks': {
        'add':
            [('timeline', 'name',), ('list_id', 'parse',)],
        'addTags':
            [('timeline', 'list_id', 'taskseries_id', 'task_id', 'tags'),
             ()],
        'complete':
            [('timeline', 'list_id', 'taskseries_id', 'task_id',), ()],
        'delete':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'), ()],
        'getList':
            [(),
             ('list_id', 'filter', 'last_sync')],
        'movePriority':
            [('timeline', 'list_id', 'taskseries_id', 'task_id', 'direction'),
             ()],
        'moveTo':
            [('timeline', 'from_list_id', 'to_list_id', 'taskseries_id',
              'task_id'),
             ()],
        'postpone':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ()],
        'removeTags':
            [('timeline', 'list_id', 'taskseries_id', 'task_id', 'tags'),
             ()],
        'setDueDate':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('due', 'has_due_time', 'parse')],
        'setEstimate':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('estimate',)],
        'setLocation':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('location_id',)],
        'setName':
            [('timeline', 'list_id', 'taskseries_id', 'task_id', 'name'),
             ()],
        'setPriority':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('priority',)],
        'setRecurrence':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('repeat',)],
        'setTags':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('tags',)],
        'setURL':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ('url',)],
        'uncomplete':
            [('timeline', 'list_id', 'taskseries_id', 'task_id'),
             ()],
    },
    'tasksNotes': {
        'add':
            [('timeline', 'list_id', 'taskseries_id', 'task_id',
              'note_title', 'note_text'), ()],
        'delete':
            [('timeline', 'note_id'), ()],
        'edit':
            [('timeline', 'note_id', 'note_title', 'note_text'), ()]
    },
    'test': {
        'echo':
            [(), ()],
        'login':
            [(), ()]
    },
    'time': {
        'convert':
            [('to_timezone',), ('from_timezone', 'to_timezone', 'time')],
        'parse':
            [('text',), ('timezone', 'dateformat')]
    },
    'timelines': {
        'create':
            [(), ()]
    },
    'timezones': {
        'getList':
            [(), ()]
    },
    'transactions': {
        'undo':
            [('timeline', 'transaction_id'), ()]
    },
}


# response format
_LIST_COMMON = {
    'list': ['archived', 'deleted', 'id', 'locked', 'name',
             'position', 'smart'],
}

_TASK_COMMON = {
    'lists': {
        'list': ['id', {
            'taskseries': ['created', 'id', 'modified', 'name',
                           'source', {
                'tags': [],
                'participants': [],
                'notes': [],
                'task': ['added', 'completed', 'deleted', 'due',
                         'estimate', 'has_due_time', 'id', 'postponed',
                         'priority'],
                },
            ],
            },
        ],
    },
}

_TASK_TAGS = {
    'lists': {
        'list': ['id', {
            'taskseries': ['created', 'id', 'modified', 'name',
                           'source', {
                'tags': {
                    'tag': [],
                },
                'participants': [],
                'notes': [],
                'task': ['added', 'completed', 'deleted', 'due',
                         'estimate', 'has_due_time', 'id', 'postponed',
                         'priority'],
                },
            ],
            },
        ],
    },
}

_TASK_RRULE = {
    'lists': {
        'list': ['id', {
            'taskseries': ['created', 'id', 'modified', 'name',
                           'source', {
                'rrule': ['every'],
                'tags': [],
                'participants': [],
                'notes': [],
                'task': ['added', 'completed', 'deleted', 'due',
                         'estimate', 'has_due_time', 'id', 'postponed',
                         'priority'],
                },
            ],
            },
        ],
    },
}

_TASK_TAGS_RRULE = {
    'lists': {
        'list': ['id', {
            'taskseries': ['created', 'id', 'modified', 'name',
                           'source', {
                'rrule': ['every'],
                'tags': {
                    'tag': [],
                },
                'participants': [],
                'notes': [],
                'task': ['added', 'completed', 'deleted', 'due',
                         'estimate', 'has_due_time', 'id', 'postponed',
                         'priority'],
                },
            ],
            },
        ],
    },
}

API_RESPONSE = {
    'auth': {
        'checkToken': {
            'auth': {
                'token': [],
                'perms': [],
                'user': ['fullname', 'id', 'username'],
            },
        },
        'getFrob': {
            'frob': [],
        },
        'getToken': {
            'auth': {
                'token': [],
                'perms': [],
                'user': ['fullname', 'id', 'username'],
            },
        },
    },
    'contacts': {
        'add': {
            'contact': ['fullname', 'id', 'username'],
        },
        'delete': [],
        'getList': {
            'contacts': {
                'contact': ['fullname', 'id', 'username'],
            },
        },
    },
    'groups': {
        'add': {
            'group': ['id', 'name', {
                'contacts': [],
                },
            ],
        },
        'addContact': [],
        'delete': [],
        'getList': {
            'groups': {
                'group': ['id', 'name', {
                    'contact': ['id'],
                    },
                ],
            },
        },
        'removeContact': [],
    },
    'lists': {
        'add': _LIST_COMMON,
        'archive': _LIST_COMMON,
        'delete': _LIST_COMMON,
        'getList': {
            'lists': _LIST_COMMON,
        },
        'setDefaultList': [],
        'setName': _LIST_COMMON,
        'unarchive': _LIST_COMMON,
    },
    'locations': {
        'getList': {
            'locations': {
                'location': ['address', 'id', 'latitude', 'longitude',
                             'name', 'viewable', 'zoom'],
            },
        },
    },
    'reflection': {
        'getMethodInfo': {
            'method': [
                'name', 'needslogin', 'needssigning', 'requiredperms',
                {
                    'description': [],
                },
                {
                    'response': [],
                },
                {
                    'arguments': {
                        'artument': ['name', 'optional'],
                    },
                },
                {
                    'errors': {
                        'error': ['code', 'message'],
                    },
                },
            ],
        },
        'getMethods': {
            'methods': {
                'method': [],
            },
        },
    },
    'settings': {
        'getList': {
            'settings': {
                'dateformat': [],
                'defaultlist': [],
                'language': [],
                'timeformat': [],
                'timezone': [],
            },
        },
    },
    'tasks': {
        'add': _TASK_COMMON,
        'addTags': _TASK_TAGS,
        'complete': _TASK_COMMON,
        'delete': _TASK_COMMON,
        'getList': _TASK_COMMON,
        'movePriority': _TASK_COMMON,
        'moveTo': _TASK_TAGS,
        'postpone': _TASK_COMMON,
        'removeTags': _TASK_TAGS,
        'setDueDate': _TASK_COMMON,
        'setEstimate': _TASK_COMMON,
        'setLocation': _TASK_COMMON,
        'setName': _TASK_COMMON,
        'setPriority': _TASK_COMMON,
        'setRecurrence': _TASK_RRULE,
        'setTags': _TASK_TAGS_RRULE,
        'setURL': _TASK_COMMON,
        'uncomplete': _TASK_COMMON,
    },
    'tasksNotes': {
        'add': {
            'note': ['created', 'id', 'modified', 'title'],
        },
        'delete': [],
        'edit': {
            'note': ['created', 'id', 'modified', 'title'],
        },
    },
    'test': {
        'echo': {
            'method': [],
        },
        'login': {
            'user': ["id", {
                "username": [],
                },
            ],
        },
    },
    'time': {
        'convert': {
            'time': ['timezone'],
        },
        'parse': {
            'time': ['precision'],
        },
    },
    'timelines': {
        'create': {
            'timeline': [],
        },
    },
    'timezones': {
        'getList': {
            'timezones': {
                'timezone': ['current_offset', 'dst', 'id', 'name', 'offset'],
            },
        },
    },
    'transactions': {
        'undo': [],
    },
}



# -*- coding: utf-8 -*-

import os
import sys
from ConfigParser import RawConfigParser
from operator import attrgetter
from os.path import dirname, join as pathjoin, realpath
from nose.plugins.skip import SkipTest
from nose.tools import *

import rtm.rtm as RTM

# constant for test
KEY_TXT = pathjoin(dirname(realpath(__file__)), 'apikey.txt')
KEY_TXT_FORMAT = """
The format are(ConfigParser readable):
[rtm]
apikey: xxx
secret: yyy
token: zzz
"""


class TestRTM(object):

    message = {
        'skip': '\n\n'
            'Set your API key "%s" to run TestRTM for authentication\n'
            '%s' % (KEY_TXT, KEY_TXT_FORMAT),
        'cannot_read': '\n\n'
            'Cannot read "%s"\n'
            '%s' % (KEY_TXT, KEY_TXT_FORMAT),
        'no_item': '\n\n'
            'rtm.%s.%s: Cannot get response item, maybe there is no item\n',
    }

    def setup(self):
        if not os.access(KEY_TXT, os.R_OK):
            raise SkipTest(self.message['skip'])
        self.read_key_txt()
        self.rtm = RTM.createRTM(self.apikey, self.secret, self.token)

    def read_key_txt(self):
        try:
            c = RawConfigParser()
            c.read(KEY_TXT)
            self.apikey = c.get('rtm', 'apikey')
            self.secret = c.get('rtm', 'secret')
            self.token = c.get('rtm', 'token')
        except Exception as err:
            print >> sys.stderr, err
            raise SkipTest(self.message['cannot_read'])

    def assert_stat_ok(self, rsp):
        assert_equal(u"ok", rsp.stat)

    def assert_response(self, func, elem=None, **params):
        """assert the stat/attr(only top) of response from api"""
        api, method = func.func_name.replace('test_', '').split('_')
        elem = elem or api
        api_method = "%s.%s" % (api, method)
        rsp = attrgetter(api_method)(self.rtm)(**params)
        self.assert_stat_ok(rsp)  # response is ok

        rsp_attr = getattr(rsp, api)
        if not isinstance(rsp_attr, RTM.dottedDict):
            raise SkipTest(self.message['no_item'] % (api, method))
        attr = RTM.API_RESPONSE[api][method][elem].keys()[0]
        assert_true(hasattr(rsp_attr, attr))  # has item

    def test_auth_checkToken(self):
        params = {'auth_token': self.token}
        self.assert_response(self.test_auth_checkToken, **params)

    def test_contacts_getList(self):
        self.assert_response(self.test_contacts_getList)

    def test_groups_getList(self):
        self.assert_response(self.test_groups_getList)

    def test_lists_getList(self):
        self.assert_response(self.test_lists_getList)

    def test_locations_getList(self):
        self.assert_response(self.test_locations_getList)

    def test_settings_getList(self):
        self.assert_response(self.test_settings_getList)

    def test_tasks_getList(self):
        self.assert_response(self.test_tasks_getList, 'lists')

    def test_timezones_getList(self):
        self.assert_response(self.test_timezones_getList)



# -*- coding: utf-8 -*-

from os.path import (abspath, dirname, isdir)
import pep8

CURRENT_DIR = dirname(abspath(__file__))
BASE_DIR = dirname(CURRENT_DIR)


def test_pep8():
    arglist = [
        "--statistics",
        "--filename=*.py",
        "--show-source",
        "--repeat",
        "--exclude=SVGdraw.py",
        "--ignore=E302,E701",
        #"--show-pep8",
        #"-qq",
        #"-v",
        BASE_DIR,
    ]

    options, args = pep8.process_options(arglist)
    runner = pep8.input_file

    for path in args:
        if isdir(path):
            pep8.input_dir(path, runner=runner)
        elif not pep8.excluded(path):
            options.counters["files"] += 1
            runner(path)

    pep8.print_statistics()
    errors = pep8.get_count("E")
    warnings = pep8.get_count("W")
    message = "pep8: %d errors / %d warnings" % (errors, warnings)
    print message
    assert errors + warnings == 0, message



# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages
from os.path import join as pathjoin

VERSION = '0.3'

LONG_DESCRIPTION = ''.join([
    open(pathjoin('src', 'README')).read()])

CLASSIFIERS = [
    'Environment :: Web Environment',
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Internet',
]

INSTALL_REQUIRES = []
if sys.version_info < (2, 6):
    INSTALL_REQUIRES.append('simplejson')

setup(
    name='pyrtm',
    version=VERSION,
    description='Remember The Milk API',
    long_description=LONG_DESCRIPTION,
    classifiers=CLASSIFIERS,
    author='Sridhar Ratnakumar',
    author_email='srid@nearfar.org',
    url='http://bitbucket.org/srid/pyrtm/',
    license='MIT License',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['buildout.cfg']},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=dict(
        test=[
            'Nose',
            'pep8',
        ],
    ),
    test_suite='nose.collector',
    tests_require=['Nose', 'pep8'],
    entry_points="""
       [console_scripts]
       rtm_appsample = rtm.samples.app:main
    """,
)



#!/usr/bin/env python
# simple app

from rtm import createRTM

from Tkinter import *

def createApp(rtm):
    rspTasks = rtm.tasks.getList(filter='dueWithin:"1 week of today"')
    tasks = []
    if hasattr(rspTasks.tasks, "list") and \
       hasattr(rspTasks.tasks.list, "__getitem__"):
        for l in rspTasks.tasks.list:
            # XXX: taskseries *may* be a list
            if isinstance(l.taskseries, (list, tuple)):
                for t in l.taskseries:
                    tasks.append(t.name)
            else:
                tasks.append(l.taskseries.name)
    print tasks
    if not tasks:
        tasks.append('No tasks due within a week')

    root = Tk()
    root.title('My tasks due within a week')
    root.wm_attributes('-topmost', 1)
    root.wm_attributes('-alpha', 0.9)
    l = Label(text='\n'.join(tasks))
    l.pack()
    l.mainloop()

def test(apiKey, secret, token=None):
    rtm = createRTM(apiKey, secret, token)
    createApp(rtm)

def main():
    import sys
    try:
        api_key, secret = sys.argv[1:3]
    except ValueError:
        print >>sys.stderr, 'Usage: rtm_appsample APIKEY SECRET [TOKEN]'
    else:
        try:
            token = sys.argv[3]
        except IndexError:
            token = None
        test(api_key, secret, token)

if __name__ == '__main__':
    main()
